//
// Copyright (C) 2020 Jrg Mller, CH-5524
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation version 3 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see http://www.gnu.org/licenses/ .
//

//
// Die Grundlage fr dieses Projekt stammt von: Adrian Meyer
//   https://www.midimountain.com/delphi_midi.html
//
// Die Vorlage fr TFrmMidiFilter.OpenMidi ist von jaenicke / entwickler-ecke.de
//
// Erweiterungen sind von: juerg5524.ch / Juli 2020
//
// Diese Anwendung verwendet "VirtualMidi" von: http://www.tobias-erichsen.de
// Damit steht ein "Loop Back" (Delphi loopback) zur Verfgung.
// Bitte, beachten Sie das Copyright von Tobias Erichsen!
//
// Fr Wine: sudo apt-get install libasound2-plugins:i386
// Unter Wine braucht es "VirtualMidi" nicht. In Linux gibt es standardmssig
// ein "Midi Through Port-0"
//
unit UMidiFilter;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Midi, StdCtrls, SyncObjs, CheckLst, Vcl.ExtCtrls, MPlayer,
  UMidiParser;

type
  TFrmMidiFilter = class(TForm)
    memLog: TMemo;
    Label3: TLabel;
    FileOpenDialog1: TFileOpenDialog;
    GroupBox2: TGroupBox;
    cbControlB: TCheckBox;
    GroupBox1: TGroupBox;
    cb0: TCheckBox;
    cb5: TCheckBox;
    cb1: TCheckBox;
    cb2: TCheckBox;
    cb3: TCheckBox;
    cb4: TCheckBox;
    cb6: TCheckBox;
    cb7: TCheckBox;
    cbC: TCheckBox;
    cbB: TCheckBox;
    cbA: TCheckBox;
    cb9: TCheckBox;
    cb8: TCheckBox;
    cbF: TCheckBox;
    cbE: TCheckBox;
    cbD: TCheckBox;
    GroupBox3: TGroupBox;
    MediaPlayer1: TMediaPlayer;
    btnPlayDelphi: TButton;
    btnPlayMidi: TButton;
    btnOpenFile: TButton;
    edtMidiFile: TEdit;
    cbMidiFilter: TCheckBox;
    cbGenFile: TCheckBox;
    GroupBox4: TGroupBox;
    Label1: TLabel;
    lbxInputDevices: TCheckListBox;
    lbxOutputDevices: TCheckListBox;
    Label2: TLabel;
    btnRefresh: TButton;
    cbControlD: TCheckBox;
    GroupBox5: TGroupBox;
    btnPiano: TButton;
    btnOergeli: TButton;
    cbInstrument: TComboBox;
    cbChannel: TComboBox;
    Label5: TLabel;
    Label6: TLabel;
    gbSimpleRec: TGroupBox;
    cbRecordInput: TCheckBox;
    btnPlayRec: TButton;
    btnDelete: TButton;
    btnUncheckAll: TButton;
    btnMakeMid: TButton;
    btnRecordMid: TButton;
    Label4: TLabel;
    btnSwissPure1: TButton;
    Label7: TLabel;
    cbSwissPure1: TComboBox;
    cbSwissPure2: TComboBox;
    btnSwissPure2: TButton;
    cbSwissPure3: TComboBox;
    btnSwissPure3: TButton;
    cbControlE: TCheckBox;
    btnToGriff: TButton;
    btnToSound: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure lbxInputDevicesClickCheck(Sender: TObject);
    procedure lbxOutputDevicesClickCheck(Sender: TObject);
    procedure btnPlayMidiClick(Sender: TObject);
    procedure btnOpenFileClick(Sender: TObject);
    procedure btnRecordMidClick(Sender: TObject);
    procedure btnPlayDelphiClick(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnOergeliClick(Sender: TObject);
    procedure btnPianoClick(Sender: TObject);
    procedure btnRefreshClick(Sender: TObject);
    procedure cbRecordInputClick(Sender: TObject);
    procedure btnPlayRecClick(Sender: TObject);
    procedure cbRecMidiClick(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure btnUncheckAllClick(Sender: TObject);
    procedure btnMakeMidClick(Sender: TObject);
    procedure btnSwissPure1Click(Sender: TObject);
    procedure btnToGriffClick(Sender: TObject);
  private
    fCriticalSection: TCriticalSection;
    stop: boolean;
    Header: TMidiHeader;
    RecordFile: System.Text;
    LastTime: integer;
    UsedChannel: array [0..15] of boolean;

    procedure OpenMidi(FileName: string);
    procedure SendMidi(Instr, Data1, Data2: byte); overload;
    procedure SendMidi(d: TInt4); overload;
    procedure CheckFirstOutputDev;
    procedure CheckOutputDev(Dev: integer; Check: boolean);
    procedure StopUsedChannel;
    procedure ClearUsedChannel;
    procedure ChangeInstrument(Bank, Channel, Instrument: byte);
    procedure RebuildMidiDevices;
  public
    procedure DoMidiInData( const aDeviceIndex: integer; const aStatus, aData1, aData2: byte;
      Timestamp: integer );
    procedure DoSysExData( const aDeviceIndex: integer; const aStream: TMemoryStream );
    function CannelChecked(Cannel: byte): boolean;
  end;

var
  FrmMidiFilter: TFrmMidiFilter;

implementation

{$R *.dfm}

  
const
  RecFile_ = 'MidiRecordInput';
  RecTxt  = RecFile_ + '.txt';
  RecMid  = RecFile_ + '.mid';


 
procedure TFrmMidiFilter.ClearUsedChannel;
var
  i: integer;
begin
  for i := 0 to High(UsedChannel) do
    UsedChannel[i] := false;
end;

procedure TFrmMidiFilter.StopUsedChannel;
var
  i, c: integer;
begin
   for c := 0 to High(UsedChannel) do
     if UsedChannel[c] then
       for i := 0 to 127 do
         SendMidi($80 + c, i, 0);
end;

procedure TFrmMidiFilter.btnPlayMidiClick(Sender: TObject);
begin
  if btnPlayMidi.Caption = 'Stop' then
    stop := true
  else
  if MediaPlayer1.FileName <> '' then
    Button4Click(Sender)
  else begin
    CheckFirstOutputDev;
    Application.ProcessMessages;
    try
      stop := false;
      gbSimpleRec.Visible := false;
      btnPlayMidi.Caption := 'Stop';
      OpenMidi(edtMidiFile.Text);
    finally
      btnPlayMidi.Caption := 'Play Midi File';
      gbSimpleRec.Visible := true;
      if stop then
        StopUsedChannel;
    end;
  end;
end;

//cbRecordInput.Checked

procedure TFrmMidiFilter.btnOpenFileClick(Sender: TObject);
begin
  FileOpenDialog1.FileName := edtMidiFile.Text;
  if FileOpenDialog1.Execute then
  begin
    edtMidiFile.Text := FileOpenDialog1.FileName;
    if not FileExists(FileOpenDialog1.FileName) then
      raise Exception.Create('File does not exist.');
  end;
end;

{$define _test_option}
procedure TFrmMidiFilter.btnRecordMidClick(Sender: TObject);
var
  SimpleFile: TSimpleMidiDataStream; 
  MidiFile, NewMidi: TMidiDataStream;
  FileName: string;
  pos: integer;
  to_Do, back: eTranslate;
{$if defined(test_option)}
  ErrFile: System.Text;
{$endif}
begin
{$if defined(test_option)}
  AssignFile(ErrFile, 'TestFile.txt');
  Rewrite(ErrFile);
  MidiConverterDirTest('E:/git/midifile_js/sounds/', ErrFile);
  CloseFile(ErrFile);
{$else}
  to_Do := nothing;
  back := nothing;
  if Sender = btnToSound then
  begin
    to_Do := toSound;
    back := toGriff;
  end else
  if Sender = btnToGriff then
  begin
    to_Do := toGriff;
    back := toSound;
  end;  
  FileName := RecMid;
  SimpleFile := TSimpleMidiDataStream.Create;
  MidiFile := TMidiDataStream.Create;
  NewMidi := TMidiDataStream.Create;
  try
    MidiFile.LoadFromFile(edtMidiFile.Text);
    if SimpleFile.MakeSimpleFile(MidiFile, to_do) then
      SimpleFile.SaveToFile(RecTxt);

    if NewMidi.MakeMidiFile(SimpleFile, back) then begin
      NewMidi.SaveToFile(FileName);
      pos := NewMidi.Compare(MidiFile);
      if (pos < NewMidi.Size) or (pos < MidiFile.Size) then
        writeln(Format('%x (%d)  %x  %x', [pos, pos, NewMidi.size, MidiFile.Size]));
    end;
  finally
    SimpleFile.Free;
    MidiFile.Free;
    NewMidi.Free;
  end;   
{$endif}
end;

procedure TFrmMidiFilter.btnPlayDelphiClick(Sender: TObject);
var
  i: integer;
begin
  if MediaPlayer1.FileName <> '' then
  begin
    Button4Click(nil);
    gbSimpleRec.Visible := true;
    exit;
  end;

  try
    if SysUtils.FileExists(edtMidiFile.Text) then
    begin
      gbSimpleRec.Visible := false;
      for i := 0 to lbxOutputDevices.Items.Count-1 do
        CheckOutputDev(i, false);
      MidiOutput.CloseAll;
      MediaPlayer1.FileName := edtMidiFile.Text;
      MediaPlayer1.Wait := true;
      MediaPlayer1.Open;
      // MediaPlayer1.Play;
      MediaPlayer1.EnabledButtons :=
        [TMPBtnType.btPause, TMPBtnType.btStop, TMPBtnType.btPlay];
      btnPlayDelphi.Caption := 'Delphi Stop';
    end;
  except
    MediaPlayer1.FileName := '';
  end;
end;

procedure TFrmMidiFilter.btnToGriffClick(Sender: TObject);
begin
  //
end;

procedure TFrmMidiFilter.Button4Click(Sender: TObject);
begin
  if MediaPlayer1.FileName <> '' then
  begin
    MediaPlayer1.FileName := '';
    btnPlayDelphi.Caption := 'Delphi Play';
    MediaPlayer1.Stop;
    MediaPlayer1.Close;
  end;
end;

function TFrmMidiFilter.CannelChecked(Cannel: byte): boolean;
begin
  result := false;
  case Cannel and $f of
    0: if cb0.Checked then Result := true;
    1: if cb1.Checked then Result := true;
    2: if cb2.Checked then Result := true;
    3: if cb3.Checked then Result := true;
    4: if cb4.Checked then Result := true;
    5: if cb5.Checked then Result := true;
    6: if cb6.Checked then Result := true;
    7: if cb7.Checked then Result := true;
    8: if cb8.Checked then Result := true;
    9: if cb9.Checked then Result := true;
    10: if cbA.Checked then Result := true;
    11: if cbB.Checked then Result := true;
    12: if cbC.Checked then Result := true;
    13: if cbD.Checked then Result := true;
    14: if cbE.Checked then Result := true;
    15: if cbF.Checked then Result := true;
  end;
end;
procedure TFrmMidiFilter.cbRecMidiClick(Sender: TObject);
begin
  //cbRecordInput.Enabled := not cbRecMidi.Checked;
  // cbGenFile.Enabled := not cbRecMidi.Checked;
end;

procedure TFrmMidiFilter.cbRecordInputClick(Sender: TObject);
var
  SimpleMidi: TSimpleMidiDataStream;
begin
  if cbRecordInput.Checked then
  begin
    btnPlayRec.Enabled := false;
    Header.DeltaTimeTicks := 96;
    if not FileExists(RecTxt) then
    begin
      system.Rewrite(RecordFile);
      writeln(RecordFile, 'Header 1 2 96');
      writeln(RecordFile, 'Meta-Event 255 81 3 7 161 32');
      writeln(RecordFile, 'Meta-Event 255 89 2 253 0');
      writeln(RecordFile, 'Meta-Event 255 88 4 3 2 24 8');
      writeln(RecordFile, 'Meta-Event 255 47 0');
      writeln(RecordFile, 'New Track 0');
      writeln(RecordFile, '0 176 0 0');
      writeln(RecordFile, '0 176 32 0');
      writeln(RecordFile, '0 192 16 0');
      writeln(RecordFile, '192 176 7 100');
    end else begin
      SimpleMidi := TSimpleMidiDataStream.Create;
      try
        SimpleMidi.LoadFromFile(RecTxt);
        if SimpleMidi.ReadSimpleHeader then
          Header := SimpleMidi.MidiHeader;
      finally
        SimpleMidi.Free;
      end;
      system.Append(RecordFile);
    end;
  end else begin
{$I-}  
    system.CloseFile(RecordFile);
{$I+}
    btnPlayRec.Enabled := true;
  end;
end;

var 
  oldtimestamp: integer = -1;
  
procedure TFrmMidiFilter.DoMidiInData(const aDeviceIndex: integer; const aStatus,
  aData1, aData2: byte; Timestamp: integer);
var
  time: double;
  NewTime: integer;
  DeltaTime: integer;
  s: string;
begin
  // skip active sensing signals from keyboard
  if aStatus = $FE then Exit;

  fCriticalSection.Acquire;
  try
    case aStatus and $f0 of
      $80,
      $90: if CannelChecked(aStatus) then exit;
      $b0: if cbControlB.Checked then exit;
      $d0: if cbControlD.Checked then exit;
      $e0: if cbControlE.Checked then exit;

      else begin end;
    end;

    if oldtimestamp < 0 then
      oldtimestamp := timestamp;
    DeltaTime := timestamp - oldtimestamp;
    if DeltaTime < 0 then
      DeltaTime := 0;
    
    DeltaTime := trunc(DeltaTime*Header.DeltaTimeTicks / (2*192.0));

    time := now;
    NewTime := round(24*3600*1000*(time - trunc(time)));
    if (LastTime <= 0) then
      LastTime := NewTime;
    if NewTime - LastTime > 2000 then
      LastTime := NewTime;   

    // send the messages on to the selected output devices
    SendMidi(aStatus, aData1, aData2);
   
    if cbRecordInput.Checked then
      writeln(RecordFile, Format('%d %d %d %d', [DeltaTime, aStatus, aData1, aData2 ]));
    s :=  Format( '%s [ %s  %d ]: %2.2x %2.2x %2.2x   %d  %d  %d',
                  [ MidiInput.DeviceNames[aDeviceIndex],
                    FormatDateTime( 'HH:NN:SS.ZZZ', time ),
                    DeltaTime,
                    aStatus,
                    aData1,
                    aData2,
                    aStatus,
                    aData1,
                    aData2]);
    if (aStatus shr 4) in [8, 9] then
      s := s +  MidiNote(aData1);
    memLog.Lines.Append(s);
    LastTime := NewTime;
    oldtimestamp := Timestamp;
  finally                                                  
    fCriticalSection.Leave;
  end;
end;

procedure TFrmMidiFilter.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if cbRecordInput.Checked then
    system.CloseFile(RecordFile);

  MidiInput.CloseAll;
end;

procedure TFrmMidiFilter.RebuildMidiDevices;
begin
  MidiInput.GenerateList;
  MidiOutput.GenerateList;

  lbxInputDevices.Items.Assign( MidiInput.DeviceNames );
  lbxOutputDevices.Items.Assign( MidiOutput.DeviceNames );  
end;

procedure TFrmMidiFilter.FormCreate(Sender: TObject);
begin
  fCriticalSection := TCriticalSection.Create;

  RebuildMidiDevices;
  MidiInput.OnMidiData := DoMidiInData;
  MidiInput.OnSysExData := DoSysExData;  

  Header.DeltaTimeTicks := 192;

  AssignFile(RecordFile, RecTxt);
end;

procedure TFrmMidiFilter.FormDestroy(Sender: TObject);
begin
  FreeAndNil( fCriticalSection );
end;

procedure TFrmMidiFilter.CheckOutputDev(Dev: integer; Check: boolean);
begin
  lbxOutputDevices.ItemIndex := Dev;
  lbxOutputDevices.Checked[Dev] := Check;
  lbxOutputDevicesClickCheck(self);
end;

procedure TFrmMidiFilter.CheckFirstOutputDev;
var
  i: integer;
  IsChecked: boolean;
begin
  IsChecked := false;
  for i:= MidiOutput.DeviceNames.Count - 1 downto 0 do
    if lbxOUtputDevices.Checked[i] then
      IsChecked := true;

  if not IsChecked and (MidiOutput.DeviceNames.Count > 0) then
    CheckOutputDev(0, true);
end;

procedure TFrmMidiFilter.lbxInputDevicesClickCheck(Sender: TObject);
begin
  if lbxInputDevices.Checked[ lbxInputDevices.ItemIndex ] then
    MidiInput.Open( lbxInputDevices.ItemIndex )
  else
    MidiInput.Close( lbxInputDevices.ItemIndex )
end;

procedure TFrmMidiFilter.lbxOutputDevicesClickCheck(Sender: TObject);
begin
  if lbxOutputDevices.Checked[ lbxOutputDevices.ItemIndex ] then
    MidiOutput.Open( lbxOutputDevices.ItemIndex )
  else
    MidiOutput.Close( lbxOutputDevices.ItemIndex )
end;

procedure TFrmMidiFilter.DoSysExData(const aDeviceIndex: integer;
  const aStream: TMemoryStream);
begin
  fCriticalSection.Acquire;
  try
    // print the message log
    memLog.Lines.Append(Format( '[%s] %s: <Bytes> %d <SysEx> %s', [ FormatDateTime( 'HH:NN:SS.ZZZ', now ),
        MidiInput.DeviceNames[aDeviceIndex],
        aStream.Size,
        SysExStreamToStr( aStream ) ] ));
  finally
    fCriticalSection.Leave;
  end;
end;


procedure TFrmMidiFilter.btnDeleteClick(Sender: TObject);
begin
  if SysUtils.FileExists(RecTxt) then
  begin
    if cbRecordInput.Checked then
      cbRecordInput.Checked := false;
    if not SysUtils.DeleteFile(RecTxt) then
      writeln('File not deleted!');;
  end;
end;

procedure TFrmMidiFilter.btnMakeMidClick(Sender: TObject);
var
  SimpleFile: TSimpleMidiDataStream;
  MidiFile: TMidiDataStream;
 begin
  SimpleFile := TSimpleMidiDataStream.Create;
  MidiFile := TMidiDataStream.Create;
  try
    SimpleFile.LoadFromFile(RecTxt);
      
    if MidiFile.MakeMidiFile(SimpleFile) then 
      MidiFile.SaveToFile(RecMid);
  finally
    SimpleFile.Free;
    MidiFile.Free;
  end;   
end;

procedure TFrmMidiFilter.ChangeInstrument(Bank, Channel, Instrument: byte);
begin
  SendMidi($b0, $00, Bank); 
  SendMidi($c0 or Channel, Instrument, $00);
  memLog.Lines.Append(Format('B0 00 %2.2x', [Bank]));
  memLog.Lines.Append(Format('%2.2x %2.2x', [$c0 or Channel, Instrument]));
end;

procedure TFrmMidiFilter.btnOergeliClick(Sender: TObject);
begin
  CheckFirstOutputDev;
  ChangeInstrument(0, cbChannel.ItemIndex, $15);
{$I-}
  if cbRecordInput.Checked then
    system.Writeln(RecordFile, Format('0 %d %d 0', [$c0  or cbChannel.ItemIndex, $15]));
{$I+}
end;

procedure TFrmMidiFilter.btnPianoClick(Sender: TObject);
begin
  CheckFirstOutputDev;
  ChangeInstrument(0, cbChannel.ItemIndex, cbInstrument.ItemIndex);
{$I-}
  if cbRecordInput.Checked then
    system.Writeln(RecordFile, Format('0 %d %d 0', [$c0  or cbChannel.ItemIndex, cbInstrument.ItemIndex]));
{$I+}
end;

procedure TFrmMidiFilter.btnSwissPure1Click(Sender: TObject);
var
  Bank, Index: byte;
  i: integer;
  Ok: boolean;
begin
  Ok := false;
  for i := 0 to lbxOutputDevices.Items.Count-1 do
   if lbxOutputDevices.Checked[i] then
     Ok := true;
  if not Ok then                                                                                
  begin  
    ShowMessage('Es ist kein "Output" gewhlt!');    
    exit;
  end;
  
  if Sender = btnSwissPure1 then
  begin
    Bank := 64;
    Index := cbSwissPure1.ItemIndex;
  end else
  if Sender = btnSwissPure2 then
  begin
    Bank := 66;
    Index := cbSwissPure2.ItemIndex;
  end else begin
    Bank := 70;
    Index := cbSwissPure3.ItemIndex;  
  end;
  ChangeInstrument(Bank-1, cbChannel.ItemIndex, Index);
end;

procedure TFrmMidiFilter.btnPlayRecClick(Sender: TObject);
var
  d: TInt4;
  Ok: boolean;
  t, t1: integer;
  SimpleMidi: TSimpleMidiDataStream;
begin
  if btnPlayRec.Caption = 'Stop' then
  begin
    stop := true;
    exit;
  end;

  CheckFirstOutputDev;
  ClearUsedChannel;
  memLog.Lines.Clear;
  Application.ProcessMessages;
  SimpleMidi := TSimpleMidiDataStream.Create;
  try
    btnPlayRec.Caption := 'Stop';
    btnPlayMidi.Enabled := false;
    btnPlayDelphi.Enabled := false;
    stop := false;
    SimpleMidi.LoadFromFile(RecTxt);
    if not SimpleMidi.ReadSimpleHeader then
      exit;
    
    while not SimpleMidi.EOF and not stop do
    begin
      Ok := SimpleMidi.ReadSimpleMidiEvent(d);
      SimpleMidi.ReadLine;
      if not Ok then
        continue;
      SendMidi(d);
      t := d[0];
      if t > 2000 then
        t := 2000;
      t1 := t;
      if (d[1] and $f0) = $90 then
        UsedChannel[d[1] and $f] := true;
      memLog.Lines.Append(Format('%d  %2.2x %2.2x %2.2x %s', [t1, d[1], d[2], d[3], MidiNote(d[2])]));
      if (t > 0) then begin
        SimpleMidi.MidiWait(t);          
        Application.ProcessMessages;
      end;
    end;
  finally
    StopUsedChannel;
    btnPlayMidi.Enabled := true;
    btnPlayDelphi.Enabled := true;
    btnPlayRec.Caption := 'Play Record File';
    SimpleMidi.Free;
  end;
end;

procedure TFrmMidiFilter.btnRefreshClick(Sender: TObject);
begin
  RebuildMidiDevices;
  Application.ProcessMessages;
end;

procedure TFrmMidiFilter.btnUncheckAllClick(Sender: TObject);
var
  Check: boolean;
begin
  check := BtnUncheckAll.Caption <> 'Uncheck All';
    
  cb0.Checked := check;
  cb1.Checked := check;
  cb2.Checked := check;
  cb3.Checked := check;
  cb4.Checked := check;
  cb5.Checked := check;
  cb6.Checked := check;
  cb7.Checked := check;
  cb8.Checked := check;
  cb9.Checked := check;
  cbA.Checked := check;
  cbB.Checked := check;
  cbC.Checked := check;
  cbD.Checked := check;
  cbE.Checked := check;
  cbF.Checked := check;

  cbControlB.Checked := check;
  cbControlD.Checked := check;
  cbControlE.Checked := check;

  if check then
    BtnUncheckAll.Caption := 'Uncheck All'
  else
    BtnUncheckAll.Caption := 'Check All';
end;

procedure TFrmMidiFilter.OpenMidi(FileName: string);
// Vorlage aus: entwickler-ecke.de (jaenicke)
var
  MidiFile: TMidiDataStream;
  GenFile: boolean;
  RecMidi: boolean;
  midi_text_file: System.Text;

  procedure ShowMsg(const msg : string);
  begin
    if GenFile then
      writeln(midi_text_file, msg);
    memLog.Lines.Append(msg);
  end;

  procedure ReadEnd;
  var 
    b: byte;
  begin
    if RecMidi then
    begin
      repeat
        if MidiFile.ChunkSize = 0 then
          break;
        b := MidiFile.NextByte;
        if (b < $80) then
          write(RecordFile, Format(' %d', [MidiFile.ReadByte]));
      until b >= $80;
      writeln(RecordFile);
    end;        
  end;

type
  TMidiFileFormat = (mffSingleTrack = 0, mffMultiTrackSync = 1, mffMultiTrackAsync = 2);
var
  TrackHeader: TMidiTrackHeader;
  com: Byte;
  Event: byte;
  Data1, Data2: byte;
  VarLen: cardinal;
  i: integer;
  b: byte;
begin
  GenFile := cbGenFile.Checked;
  if GenFile then
  begin
    AssignFile(midi_text_file, 'Midi.txt');
    system.Rewrite(midi_text_file);
  end;
  MidiFile := TMidiDataStream.Create;
  MidiFile.LoadFromFile(FileName);

  RecMidi := cbRecordInput.Checked and cbRecordInput.Enabled;
  if RecMidi then
    system.Rewrite(RecordFile);

  memLog.Clear;
  try
    cbGenFile.Enabled := false;
    gbSimpleRec.Enabled := false;
    
    if not MidiFile.ReadMidiHeader(true) then
      exit;

    with MidiFile.MidiHeader do
    begin
      ShowMsg('Format: ' + IntToStr(ord(FileFormat)) + '  Track Count: ' + IntToStr(TrackCount) +
              '  Time Ticks: ' + IntToStr(DeltaTimeTicks));

      if RecMidi then
        writeln(RecordFile, cSimpleHeader + ' ' + IntToStr(ord(FileFormat)) + ' ' + IntToStr(TrackCount) + ' ' + IntToStr(DeltaTimeTicks));
    end;
    // ### Tracks ###
    stop := false;
    while not stop and ( MidiFile.Position < MidiFile.Size) do
    begin
      if not MidiFile.ReadMidiTrackHeader(TrackHeader, true) then
        exit;
        
      ShowMsg('Delta Time: ' + IntToStr(TrackHeader.DeltaTime));
      // Wait(DeltaTime);

      if RecMidi then
        writeln(RecordFile, cSimpleTrackHeader + ' ' + IntToStr(TrackHeader.DeltaTime));
      
      while MidiFile.ChunkSize > 0 do
      begin
        if stop then                                              
          break;

        com := MidiFile.ReadByte;
        while (com < $80) and (MidiFile.ChunkSize > 0) do
          com := MidiFile.ReadByte;
        if MidiFile.ChunkSize = 0 then
          break;
          
        Event := com shr 4;
        Data1 := MidiFile.ReadByte;
        Data2 := 0;
        if ((com shr 4) in [$8..$c, $e]) and
           (MidiFile.NextByte < $80) then
          Data2 := MidiFile.ReadByte;
        VarLen := 0;
        case com shr 4 of
          $8:
            begin
              //inc(Data1, 5);
 //             ShowMsg('Note off for Channel ' + IntToStr(EventChannel)
 //               + ' - Data1: ' + IntToHex(Data1) + ', Data2: ' + IntToHex(VarLen));
            end;
          $9:
            begin
              //inc(Data1, 5);
              UsedChannel[com and $f] := true;
  //            ShowMsg('Note on for Channel ' + IntToStr(EventChannel)
    //            + ' - Data1: ' + IntToHex(Data1) + ', Data2: ' + IntToHex(VarLen));
            end;
          $A:
            begin
              ShowMsg('Key after-touch for Channel ' + IntToStr(com and $f)
                + ' - Data1: ' + IntToStr(Data1) + ', Data2: ' + IntToStr(Data2));
            end;
          $B:
            begin
              VarLen := MidiFile.ReadVariableLen;
              ShowMsg('Control Change for Channel ' + IntToStr((com and $f))
                + ' - Data1: ' + IntToStr(Data1) + ', Data2: ' + IntToStr(Data2) + ' ' + IntToStr(VarLen));
            end;
          $C:
            begin
              ShowMsg('Program (patch) change for Channel ' + IntToStr((com and $f))
                + ' - Data: ' + IntToStr(Data1));
            end;
          $D:
            begin
              ShowMsg('Channel after-touch for Channel ' + IntToStr(com and $f)
                + ' - Data: ' + IntToStr(Data1));
            end;
          $E:
            begin
              ShowMsg('Pitch wheel change for Channel ' + IntToStr((com and $f))
                + ' - Data1: ' + IntToStr(Data1) + ', Data2: ' + IntToStr(Data2));
            end;
          $F:
            begin
              // Meta-Event
              VarLen := MidiFile.ReadVariableLen;
              if RecMidi then
                write(RecordFile, 'Meta-Event ' + IntToStr(com) + ' ' +
                  IntToStr(Data1) + ' ' + IntToStr(VarLen));
              ShowMsg('Meta-Event Length ' + IntToStr(VarLen));
              for i := 1 to VarLen do begin
                b := MidiFile.ReadByte;
                if RecMidi then
                  write(RecordFile, ' ' + IntToStr(b));
                //ShowMsg('  ' + IntToStr(i) + ':  ' + IntToHex(b));
              end;
              if VarLen > 0 then 
              begin
                VarLen := MidiFile.ReadVariableLen;
                if RecMidi then
                begin
                  write(RecordFile, ' ' + IntToStr(VarLen));
                end;
              end;
              ReadEnd;
            end;
            else begin
            end;
        end;
        if Event in [8..14] then
        begin
          if Event in [$8, $9, $a, $e] then
            VarLen := MidiFile.ReadVariableLen;
          if GenFile then
            writeln(midi_text_file, Format('%2.2x %2.2x %2.2x %x', [com, Data1, Data2, VarLen]));
          if RecMidi then
          begin
            write(RecordFile, Format('%d %d %d %d', [VarLen, com, Data1, Data2]));
            ReadEnd;
          end else begin
            SendMidi(com, Data1, Data2);
            if Event in [$8, $9, $a, $e] then begin
              MidiFile.MidiWait(VarLen);
            end;
          end;
        end;
        Application.ProcessMessages;
      end;
    end;

    MidiFile.Free;
  except
    on E: Exception do
    begin
      ShowMessage('Fehler: ' + E.Message + ' an Position ' + IntToStr(MidiFile.Position) + ' $' + IntToHex(MidiFile.Position));
      MidiFile.Free;
    end;
  end;
  if GenFile then
    CloseFile(midi_text_file);
  if RecMidi then
    cbRecordInput.Checked := false;
  cbGenFile.Enabled := true;
  gbSimpleRec.Enabled := true;
end;

procedure TFrmMidiFilter.SendMidi(Instr, Data1, Data2: byte);
var
  i: integer;
begin
  // send the messages to the selected output devices
  for i:= lbxOutputDevices.Count - 1 downto 0 do
    if lbxOutputDevices.Checked[i] and
       not (cbMidiFilter.Checked and CannelChecked(Instr)) then
    begin
      MidiOutput.Send(i, Instr, Data1, Data2 );
    end;
end;

procedure TFrmMidiFilter.SendMidi(d: TInt4);
begin
  SendMidi(d[1], d[2], d[3]);
end;

end.
