//
// Copyright (C) 2020 Jrg Mller, CH-5524
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation version 3 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see http://www.gnu.org/licenses/ .
//

unit UMidiParser;

interface

uses
  Classes, SysUtils, Dialogs;

type
  TPitchArray = array [0..10] of byte; 
  TGriffArray = array [0..6] of byte; // tiefste Oktave in As-Dur
  eTranslate = (nothing, toGriff, toSound); 
    
  TVocalArray = 
    record
      Arr1: TPitchArray;
      Arr2: TPitchArray;
      Arr3: TPitchArray;
      procedure Transpose(delta: integer);      
    end;
  PVocalArray = ^TVocalArray;

  TInstrument = 
    record
      Push: TVocalArray;
      Pull: TVocalArray;
      Griff: TGriffArray;
      
      function GriffToSound(Pitch: byte; const VocalArr: TVocalArray; Cross: boolean) : integer; overload;
      function SoundToGriff(const Pitch: byte; const VocalArr: TVocalArray; Arr: pinteger = nil): integer; overload;
      function GriffToSound(Pitch: byte; Push: boolean; Cross: boolean) : integer; overload;
      function SoundToGriff(const Pitch: byte; Push: boolean; Arr: pinteger = nil): integer; overload;
      function IndexToGriff(index: byte): byte;
      function GetPitchLine(pitch: byte): integer;
      procedure Transpose(delta: integer);
    end;
    
const
  cSimpleHeader = 'Header';
  cSimpleTrackHeader = 'New Track';

  PushTest = true;
  CrossTest = true;

  HexOutput = true;

  MidiC0 = 12;
  Notes : array [0..11] of string = ('C', 'Des', 'D', 'Es', 'E', 'F', 'Ges', 'G', 'As', 'A', 'B', 'H');

  Gwerder_b_Oergeli : TInstrument = (
    Push : (    
      // Oben -> Unten [absolute height of the note (MIDI).Tonal Pitch Classes] 
      //   Schwyzerrgeli normal, 31 Tasten, A/D/G/C                      B4
      Arr1 : (50,53,58,62,65,70,74,77,82,86,00);  // D3   F3 B3   D4  F4  B4   D5  F5 B5   D6
      Arr2 : (64,55,58,63,67,70,75,79,82,87,91);  // Es4  G3 B3   Es4 G4  B4   Es5 G5 B5   Es6 G6
      Arr3 : (54,71,61,72,68,73,80,84,85,83,00);  // Ges3 H4 Des4 C5  As4 Des5 As5 C6 Des6 H5
    );                                            

    Pull: (                               
      Arr1 : (53,57,60,63,67,69,72,75,79,81,00);                             
      Arr2 : (64,58,62,65,68,72,74,77,80,84,86);
      Arr3 : (49,71,56,66,73,70,78,82,85,88,00);
    );
    // As-Dur-Noten
    //         f,  g, as,  b,  c,des, es    
    Griff : ( 53, 55, 56, 58, 60, 61, 63 ); 
  );          


type
  TInt4 = array [0..3] of integer;
  
  TMidiHeader = record
    FileFormat: word;
    TrackCount: word;
    // delta-time ticks pro Viertelnote
    DeltaTimeTicks: word;
  end; 

  TMidiTrackHeader = record
    ChunkSize: cardinal;
    DeltaTime: cardinal;
  end;

  TSimpleTrackHeader = record
    DeltaTime: cardinal;
  end;
   
  TSimpleMidiDataStream = class;
   
  TMyDataStream = class(TMemoryStream)
  public
    MidiHeader: TMidiHeader;
    ChunkSize: Cardinal;
  
    function NextByte: byte;
    function ReadByte: byte;
    function GetByte(const Offset: cardinal): byte;
    function ReadWord: word;
    function GetWord(const Offset: cardinal): word;
    function ReadCardinal: cardinal;
    function GetCardinal(const Offset: cardinal): cardinal;        

    procedure WriteByte(b: byte); 
    procedure SetByte(b: byte; const Offset: cardinal); 
    procedure WriteWord(w: Word); 
    procedure SetWord(w: word; const Offset: cardinal); 
    procedure WriteCardinal(c: cardinal); 
    procedure SetCardinal(c: cardinal; const Offset: cardinal);
    procedure WriteString(s: AnsiString); overload;
    procedure WriteString(s: String); overload;
    procedure WritelnString(s: AnsiString); overload;
    procedure WritelnString(s: String); overload;

    procedure SkipBytes(c: cardinal);
    class function IsEndOfTrack(const d: TInt4): boolean;
    procedure MidiWait(Delay: integer);


    function Compare(Stream: TMyDataStream): integer;
  end;

  // Zur Analyse von Midi-Files (.mid).
  // Zur Generierung von Midi-Files aus einem einfachen Text-File (simple file).
  TMidiDataStream = class(TMyDataStream)
  public
    function ReadVariableLen: cardinal;
    procedure WriteVariableLen(c: cardinal);
  
    function ReadMidiHeader(RaiseExcept: boolean = false): boolean;
    function ReadMidiTrackHeader(var Header: TMidiTrackHeader; RaiseExcept: boolean = false): boolean;
    function MakeMidiFile(SimpleFile: TSimpleMidiDataStream; toDo: eTranslate = nothing): boolean; 
  end;

  // Zur Analyse von einfachen Text-Files (.txt).
  // Zur Generierung von einfachen Text-Files aus Midi-Files.
   
  TSimpleMidiDataStream = class(TMyDataStream)
  public
    function NextIsNumber: boolean;
    procedure ReadLine;
    function NextNumber: integer;
    function ReadNumber: integer;
    function EOF: boolean;

    function ReadSimpleHeader: boolean;
    function ReadSimpleTrackHeader(var TrackHeader: TSimpleTrackHeader): boolean;
    function ReadSimpleMidiEvent(var d: TInt4): boolean ;
    function MakeSimpleFile(MidiFile: TMidiDataStream; toDo: eTranslate = nothing): boolean;

  end;
  
  procedure MidiConverterTest(const FileName: string; var Text: System.Text);
  procedure MidiConverterDirTest(const DirName: string; var Text: System.Text);
  function MidiNote(Pitch: byte): string;

var
  Gwerder_a_Oergeli : TInstrument;

  
implementation

procedure TVocalArray.Transpose(delta: integer);
var i: integer;
begin
  for i := Low(TPitchArray) to High(TPitchArray)-1 do
  begin
    inc(Arr1[i], delta); 
    inc(Arr2[i], delta);
    inc(Arr3[i], delta); 
  end;
  inc(Arr2[High(TPitchArray)], delta); 
end;

procedure TInstrument.Transpose(delta: integer);
var
  i: integer;
begin
  Push.Transpose(delta);
  Pull.Transpose(delta);
  for i := Low(Griff) to High(Griff) do
    inc(Griff[i], delta);
end;


function GetIndexToPitchInArray(pitch: byte; const arr: TPitchArray): integer;
begin
  result := High(arr);
  while result >= 0 do
  begin 
    if pitch = arr[result] then
      break;
    dec(result);
  end;
end;

function TInstrument.GriffToSound(Pitch: byte; Push: boolean; Cross: boolean): integer;
begin

  if Push then
    result := GriffToSound(Pitch, self.Push, Cross)
  else
    result := GriffToSound(Pitch, self.Pull, Cross);
end;

function TInstrument.GriffToSound(Pitch: byte; const VocalArr: TVocalArray; Cross: boolean) : integer;
var
  Line, index: integer;
begin
  result := -1;
    
  Line := GetPitchLine(Pitch);
  index := Line div 2;
  if (index > High(VocalArr.Arr2)) or (Pitch = 0) then
    exit;
    
  if not odd(Line) then
    result := VocalArr.Arr2[index]
  else
  if Cross then
    result := VocalArr.Arr3[index]
  else
    result := VocalArr.Arr1[index];
  writeln(Format('%d   %d', [Pitch, result]));
end;

function TInstrument.SoundToGriff(const Pitch: byte; Push: boolean; Arr: pinteger): integer;
begin
  if Push then
    result := SoundToGriff(Pitch, self.Push, Arr)
  else
    result := SoundToGriff(Pitch, self.Pull, Arr);
end;

function TInstrument.IndexToGriff(index: byte): byte;
begin
  result := Griff[index mod 7] + 12*(index div 7);
end;

function TInstrument.GetPitchLine(pitch: byte): integer;
begin
  //            f,  g, as,  b,  c,des, es      
  // Griff : ( 53, 54, 56, 58, 60, 61, 63 ); 
  result := 0;
  while pitch > Griff[6] do
  begin
    inc(result, 7);
    dec(pitch, 12);
  end;
  dec(pitch, Griff[0]);

  case pitch of
    0: begin end;           // f
    1, 2: inc(result);      // ges, g
    3, 4: inc(result, 2);   // as, a
    5, 6: inc(result, 3);   // b, h
    7: inc(result, 4);      // c
    8, 9: inc(result, 5);   // des, d
    10, 11: inc(result, 6); // es, e
    else begin end;      
  end;
end;

function TInstrument.SoundToGriff(const Pitch: byte; const VocalArr: TVocalArray; Arr: pinteger): integer; 
var
  index: integer;
begin
  if Arr <> nil then
    Arr^ := 0;
  result := -1;
    
  index := GetIndexToPitchInArray(Pitch, VocalArr.Arr2);
  if (index >= 0) then
  begin
    result := IndexToGriff(2*index);
    if Arr <> nil then
      Arr^ := 2;
    exit;
  end;
  
  index := GetIndexToPitchInArray(Pitch, VocalArr.Arr1);
  if (index >= 0) then
  begin
    result := IndexToGriff(2*index + 1);
    if Arr <> nil then
      Arr^ := 1;
    exit;
  end;
 
  index := GetIndexToPitchInArray(Pitch, VocalArr.Arr3);
  if (index >= 0) then
  begin
    result := IndexToGriff(2*index + 1);
    if Arr <> nil then
      Arr^ := 3;
    exit;
  end;
end;      


////////////////////////////////////////////////////////////////////////////////

function TMyDataStream.NextByte: byte;
begin
  if Position < Size then
    result := GetByte(Position)
  else
    result := 0;
end;

function TMyDataStream.ReadByte: byte;
begin
  result := GetByte(Position);
  Position := Position + 1;
  if ChunkSize > 0 then
    dec(ChunkSize);
end;

function TMyDataStream.GetByte(const Offset: cardinal): byte;
begin
  if Offset < Size then
    result := PByte(Memory)[Offset]
  else
    result := 0;      
end;

function TMyDataStream.ReadWord: word;
begin
  Result := ReadByte shl 8 + ReadByte;
end;

function TMyDataStream.GetWord(const Offset: cardinal): word;
begin
  result := GetByte(Offset) shl 8 + GetByte(Offset + 1);
end;

function TMyDataStream.ReadCardinal: cardinal;
begin
  Result := ReadByte shl 24 + ReadByte shl 16 + ReadByte shl 8 + ReadByte;
end;

function TMyDataStream.GetCardinal(const Offset: cardinal): cardinal;
begin
  result := GetWord(Offset) shl 16 + GetWord(Offset + 2);
end;  

procedure TMyDataStream.WriteByte(b: byte); 
begin
  write(b, 1);
end;

procedure TMyDataStream.SetByte(b: byte; const Offset: cardinal); 
var
  old: cardinal;
begin
  old := Position;
  Position := offset;
  WriteByte(b);
  Position := Old;
end;

procedure TMyDataStream.WriteWord(w: Word); 
begin
  WriteByte(w shr 8);
  WriteByte(w and $ff);
end;

procedure TMyDataStream.SetWord(w: word; const Offset: cardinal); 
begin
  SetByte(w shr 8, Offset);
  SetByte(w and $ff, Offset + 1);
end;

procedure TMyDataStream.WriteCardinal(c: cardinal);
begin
  WriteWord(c shr 16);
  WriteWord(c and $ffff);
end;

procedure TMyDataStream.SetCardinal(c: cardinal; const Offset: cardinal);
begin
  SetWord(c shr 16, Offset);
  SetWord(c and $ffff, Offset + 2);
end;

procedure TMyDataStream.WriteString(s: AnsiString);
var
  i: integer;
begin
  for i := 1 to length(s) do 
    WriteByte(Byte(s[i]));    
end;

procedure TMyDataStream.WriteString(s: String);
begin
  WriteString(AnsiString(s));  
end;

procedure TMyDataStream.WritelnString(s: AnsiString);
begin
  WriteString(s);
  WriteString(#10); // #10#13
end;

procedure TMyDataStream.WritelnString(s: String);
begin
  WritelnString(AnsiString(s));  
end;

procedure TMyDataStream.SkipBytes(c: cardinal);
begin
  if Position + c > Size then
    Position := Size
  else
    Position := Position + c
end;

procedure TMyDataStream.MidiWait(Delay: integer);
begin
  if (Delay > 0) and (MidiHeader.DeltaTimeTicks > 0) then
  begin
    Delay := trunc(2*Delay*192.0 / MidiHeader.DeltaTimeTicks);
    if Delay > 2000 then
      Delay := 1000;
    if Delay > 16 then
      dec(Delay, 16)
    else
      Delay := 1;  
    
    Sleep(Delay);
  end;
end;

function TMyDataStream.Compare(Stream: TMyDataStream): integer;
var
  b1, b2: byte;
  Err: integer;
begin
  result := 0;
  Err := 0;
  repeat
    if (result >= Size) and (result >= Stream.Size) then
      break;
    b1:= GetByte(result);
    b2 := Stream.GetByte(result);
    if (b1 <> b2) then
    begin
      system.writeln(Format('%x (%d): %d   %d', [result, result, b1, b2]));
      inc(Err);
    end;
     // break;
    inc(result);
  until false;
  system.writeln('Err: ', Err);
end;


////////////////////////////////////////////////////////////////////////////////

function TSimpleMidiDataStream.NextIsNumber: boolean;
begin
  while (Position < Size) and (NextByte = ord(' ')) do
    ReadByte;
  result := NextByte in [ord('0')..ord('9'), ord('$')];
end;

function TSimpleMidiDataStream.NextNumber: integer;
var
  pos: cardinal;
begin
  pos := Position;
  result := ReadNumber;
  Position := pos;
end;

function TSimpleMidiDataStream.ReadNumber: integer;
var
  b: byte;
  factor: integer;
begin
  while (Position < Size) and (NextByte = ord(' ')) do
    ReadByte;

  factor := 10;
  b := NextByte;
  if b = ord('$') then
  begin
    factor := 16;
    ReadByte
  end else
  if (b < ord('0')) or (b > ord('9')) then
  begin 
    result := -1;
    exit;
  end;
  result := 0;

  repeat
    if NextByte in [ord('0')..ord('9')] then
      result := factor*result + ReadByte - ord('0')
    else
    if NextByte in [ord('A')..ord('F')] then
      result := factor*result + ReadByte - ord('A') + 10
    else
      break;
  until (false);         
end;   

procedure TSimpleMidiDataStream.ReadLine;
begin
  while NextByte >= ord(' ') do
    ReadByte;
  while (NextByte <> 0) and (NextByte <= ord(' ')) do   // skip empty lines
  begin
    while NextByte in [1..ord(' ')] do
      ReadByte;  
  end;
end;

function TSimpleMidiDataStream.EOF: boolean;
begin
  result := (Position >= Size) or (NextByte = 0);
end;

function TSimpleMidiDataStream.MakeSimpleFile(MidiFile: TMidiDataStream; toDo: eTranslate): boolean; 
var
  TrackHeader: TMidiTrackHeader;
  VarLen, i: integer;
  Com, d1, d2, b, old_d1: byte;
begin
  result := false;
  SetSize(10000000);
  Position := 0;
  MidiFile.Position := 0;

  try
    if not MidiFile.ReadMidiHeader(true) then
      exit;
      
    with MidiFile.MidiHeader do
      WritelnString(cSimpleHeader + ' ' + IntToStr(ord(FileFormat)) + ' ' + IntToStr(TrackCount) + ' ' + IntToStr(DeltaTimeTicks));

    // ### Tracks ###
    while MidiFile.Position < MidiFile.Size do
    begin
      if not MidiFile.ReadMidiTrackHeader(TrackHeader, true) then
        exit;

      WritelnString(cSimpleTrackHeader + ' ' + IntToStr(TrackHeader.DeltaTime));
      
      while MidiFile.ChunkSize > 0 do
      begin
        // Command byte
        com := MidiFile.ReadByte;
        while (com < $80) and (MidiFile.ChunkSize > 0) do
          com := MidiFile.ReadByte;
        if MidiFile.ChunkSize = 0 then
          break;
        d1 := MidiFile.ReadByte;
        old_d1 := d1;
        d2 := 0;
        VarLen := 0;
        if ((com shr 4) in [$8..$c, $e]) and (MidiFile.NextByte < $80) then
          d2 := MidiFile.ReadByte;
        if (com shr 4) in [$8, $9, $a, $b, $e, $f] then
          VarLen := MidiFile.ReadVariableLen;
        if ((com shr 4) in [8,9]) and (toDo <> nothing) then
        begin
          if toDo = toSound then
            i := Gwerder_b_Oergeli.GriffToSound(d1, PushTest, CrossTest)
          else
            i := Gwerder_b_Oergeli.GriffToSound(d1, PushTest, CrossTest);
          if i > 0 then
            d1 := i;
        end;
        case com shr 4 of
          $F: begin
              WriteString('Meta-Event ' + IntToStr(com) + ' ' + IntToStr(d1) + ' ' + IntToStr(VarLen));
              for i := 1 to VarLen do begin
                b := MidiFile.ReadByte;
                WriteString(' ' + IntToStr(b));
              end;
              if VarLen > 0 then 
              begin
                VarLen := MidiFile.ReadVariableLen;
                WriteString(' ' + IntToStr(VarLen));
              end;
            end;
          8..14: begin
              if HexOutput then
                WriteString(Format('%5d $%2.2x $%2.2x $%2.2x', [VarLen, com, d1, d2]))
              else
                WriteString(Format('%5d %3d %3d %3d', [VarLen, com, d1, d2]));
            end;
          else begin end;
        end;
        if com >= $80 then begin
          repeat
            if MidiFile.ChunkSize = 0 then
              break;
            b := MidiFile.NextByte;
            if (b < $80) then
              WriteString(Format(' %d', [MidiFile.ReadByte]));
          until b >= $80;
          if (com shr 4) in [8, 9] then
          begin
            if old_d1 <> d1 then
              WriteString(MidiNote(old_d1));
            WriteString(MidiNote(d1));
          end;
          WritelnString('');
        end;
      end;
    end;        
 
  except
    on E: Exception do
    begin
      ShowMessage('Fehler: ' + E.Message + ' an Position $' + IntToHex(MidiFile.Position));
    end;
  end;
  SetSize(Position);
  result := true;    
end;

function TSimpleMidiDataStream.ReadSimpleHeader: boolean;
begin
  result := NextByte = ord('H');
  if result then
  begin
    SkipBytes(length(cSimpleHeader));

    MidiHeader.FileFormat := ReadNumber;
    MidiHeader.TrackCount := ReadNumber;
    MidiHeader.DeltaTimeTicks := ReadNumber;
  end;
  ReadLine;
end;

function TSimpleMidiDataStream.ReadSimpleTrackHeader(var TrackHeader: TSimpleTrackHeader): boolean;
begin
  result := NextByte = ord('N');
  if result then 
  begin
    SkipBytes(length(cSimpleTrackHeader));
    TrackHeader.DeltaTime := ReadNumber;
    ReadLine;
  end;
end;

function TSimpleMidiDataStream.ReadSimpleMidiEvent(var d: TInt4): boolean;
var 
  i: integer;
begin
  result := true;
  
  d[0] := ReadNumber;
  d[1] := ReadNumber;
  d[2] := ReadNumber;
  d[3] := 0;
  if (d[1] shr 4) <> $d then
    d[3] := ReadNumber;
  for i  := Low(d) to High(d) do
    if d[i] < 0 then
      result := false;  
end;


////////////////////////////////////////////////////////////////////////////////       

function TMidiDataStream.ReadVariableLen: cardinal;
var 
  c: byte;
begin
  result := ReadByte;
  if (result and $80) <> 0 then
  begin
    result := result and $7f;
    repeat
      c := ReadByte;
      result := (result shl 7) + (c and $7f);
    until (c < $80);
  end;
end;

procedure TMidiDataStream.WriteVariableLen(c: cardinal);
var
  buffer: cardinal;
begin
  buffer := c and $7f;
  while (c shr 7) <> 0 do
  begin
    c := c shr 7;
    buffer := (buffer shl 8) + (c and $7f) + $80;
  end;
  while (true) do
  begin
    WriteByte(buffer and $ff);
    if (buffer and $80) <> 0 then
      buffer := buffer shr 8
    else
      break;
  end;
end;
  
function TMidiDataStream.MakeMidiFile(SimpleFile: TSimpleMidiDataStream; toDo: eTranslate): boolean; 
var
  d: TInt4;
  size_pos, c: cardinal;
  t: integer;
  i: integer;
  Event: byte;
  TrackHeader: TSimpleTrackHeader;
  arr: integer;
begin
  result := false;
  SetSize(SimpleFile.Size);
  SimpleFile.Position := 0;
  Position := 0;

  if not SimpleFile.ReadSimpleHeader then 
    exit;

  WriteString('MThd');
  WriteCardinal(6);     
  WriteWord(SimpleFile.MidiHeader.FileFormat);   
  WriteWord(SimpleFile.MidiHeader.TrackCount);  
  WriteWord(SimpleFile.MidiHeader.DeltaTimeTicks);  
  

  while (SimpleFile.Position + 20 < SimpleFile.Size) do
  begin
    if not SimpleFile.ReadSimpleTrackHeader(TrackHeader) then
      exit;

    // track
    WriteString('MTrk');
    size_pos := Position;
    WriteCardinal(0);
    WriteVariableLen(TrackHeader.DeltaTime);   
        
    repeat
      if (SimpleFile.NextByte = ord('N')) then
        break;
    
      if (SimpleFile.NextByte = ord('M')) then
      begin
        d[0] := 0;
        SimpleFile.SkipBytes(length('Meta-Event'));
        d[1] := SimpleFile.ReadNumber;
        WriteByte(d[1]);      
        d[2] := SimpleFile.ReadNumber;
        WriteByte(d[2]);
        d[3] := SimpleFile.ReadNumber;
        WriteVariableLen(d[3]);
        for t := 1 to d[3] do 
          WriteByte(SimpleFile.ReadNumber);
        t := SimpleFile.ReadNumber;
        if t >= 0 then
          WriteVariableLen(t);
        SimpleFile.ReadLine;
        if IsEndOfTrack(d) then
          break;
        continue;
      end;

      if not SimpleFile.ReadSimpleMidiEvent(d) then
        exit;

      Event := d[1] shr 4;
      WriteByte(d[1]);
      if (toDo <> nothing) and (Event in [8,9]) then
        if toDo = toGriff then
        begin
          if toDo = toSound then
            i := Gwerder_b_Oergeli.SoundToGriff(d[2], PushTest, @arr)
          else
            i := Gwerder_b_Oergeli.SoundToGriff(d[2], PushTest, @arr);
          system.writeln(Format('%d   %d  arr: %d', [d[2], i, arr]));
          if i > 0 then
            d[2] := i; 
        end;
      WriteByte(d[2]);
      if Event <> $d then
        WriteByte(d[3]);
      if not (Event in [$c, $d]) then begin
        WriteVariableLen(d[0]);
      end;

      repeat
        t := SimpleFile.ReadNumber; 
        if (t < 0) then
          break;
        WriteByte(t);
      until false;
      SimpleFile.Readline;
    until SimpleFile.NextByte = 0;  
    if not IsEndOfTrack(d) then
    begin
      WriteByte($ff);
      WriteByte($2f);
      WriteByte($00);
    end;      

    c := Position - size_pos - 4;
    SetCardinal(c, size_pos);
  end;

  SetSize(Position);
  result := true;    
end;

function TMidiDataStream.ReadMidiHeader(RaiseExcept: boolean): boolean;
var
  Signature, ChunkSize: cardinal;
begin
  result := false;
  
  Signature := ReadCardinal;
  if Signature <> $4D546864 then   // MThd
  begin
    if RaiseExcept then
      raise Exception.Create('Falsche Header-Signatur!');
    exit;
  end;
  ChunkSize := ReadCardinal;
  if Size - Position < ChunkSize then
  begin
    if RaiseExcept then
      raise Exception.Create('Restliche Dateigre kleiner als die angegebene Chunkgre!');
    exit;
  end;
  MidiHeader.FileFormat := ReadWord;
  MidiHeader.TrackCount := ReadWord;
  MidiHeader.DeltaTimeTicks := ReadWord;

  result := ChunkSize = 6;
end;

function TMidiDataStream.ReadMidiTrackHeader(var Header: TMidiTrackHeader; RaiseExcept: boolean = false): boolean;
var
  Signature: cardinal;
begin
  result := false;
  Signature := ReadCardinal;
  if Signature <> $4D54726B then              // MTrk
  begin
    if RaiseExcept then
      raise Exception.Create('Falsche Track-Chunk-Signatur!');
    exit;
  end;

  Header.ChunkSize := ReadCardinal;
  ChunkSize := Header.ChunkSize;
  Header.DeltaTime := ReadVariableLen;
  if (Size - Position + 1 < Header.ChunkSize) then
  begin
    if RaiseExcept then
      raise Exception.Create('Restliche Dateigre kleiner als die angegebene Chunkgre!');
    exit;
  end;
  result := true;
end;

////////////////////////////////////////////////////////////////////////////////

procedure MidiConverterTest(const FileName: string; var Text: System.Text);
var
  SimpleFile: TSimpleMidiDataStream;
  MidiFile, NewMidi: TMidiDataStream;
  pos: integer;
begin
  SimpleFile := TSimpleMidiDataStream.Create;
  MidiFile := TMidiDataStream.Create;
  NewMidi := TMidiDataStream.Create;
  try
    writeln(Text, '---> ' + Filename);
    MidiFile.LoadFromFile(FileName);
    if not SimpleFile.MakeSimpleFile(MidiFile, toSound) then
    begin
      writeln(Text, 'File not converted to simple file: ' + Filename);
      exit;
    end;

    if not NewMidi.MakeMidiFile(SimpleFile, toGriff) then begin
      writeln(Text, 'File not converted to midi file: ' + Filename);
      exit;
    end;
    pos := NewMidi.Compare(MidiFile);
    if (pos < NewMidi.Size) or (pos < MidiFile.Size) then
      writeln(Text, Format('%x (%d) %x  %x', [pos, pos, NewMidi.size, MidiFile.Size]));
  finally
    SimpleFile.Free;
    MidiFile.Free;
    NewMidi.Free;
  end;
end;

procedure MidiConverterDirTest(const DirName: string; var Text: System.Text);
var
  SR: TSearchRec;
  s: string;
begin
  if FindFirst(DirName + '*.*', faAnyFile, SR) = 0 then
  begin
    repeat
      s := DirName + SR.Name;
      if SR.Name[1] = '.' then
      else
      if SR.Attr = faDirectory then
        MidiConverterDirTest(s + '/', Text)
      else 
      if SysUtils.ExtractFileExt(SR.Name) = '.mid' then
        MidiConverterTest(s, Text);
    until FindNext(SR) <> 0;
    FindClose(SR);
  end;
end;

function MidiNote(Pitch: byte): string;
var
  code: byte;
begin
  code := Pitch;
  if Pitch >= 12 then
    dec(code, 12);
  result := Format('   %3s %d  -  %d', [Notes[code mod 12], code div 12, Pitch])
end;

class function TMyDataStream.IsEndOfTrack(const d: TInt4): boolean;
begin
  result :=  (d[1] = $ff) and (d[2] = $2f) and (d[3] = 0);
end;


initialization

  
  Gwerder_a_Oergeli := Gwerder_b_Oergeli;
  Gwerder_a_Oergeli.Transpose(+11);
  
finalization

end.


